package application;

import javafx.scene.Group;

public class MovingBall extends Group {

	// moving ball is composed of a ball and motion vector
	private Ball ball;
	private MotionVector mv;
	
	// ball size
	private double radius; 	// in pixels
	
	// physical screen size 
	private double scrWidth; // in meter
	private double scrHeight;// in meter
		
	// ball coordinates in world
	private double prevX, prevY, prevZ;
	private double currX, currY, currZ;

	// onscreen ball coordinate
	private double pixX, pixY;
	
	// motion vector coordinates in world and onscreen
	private double fromPixX, fromPixY;
	private double toPixX, toPixY;
	
	// nearest & farthest visible distance	
	private double minDepth, maxDepth; 
	
	// animation refresh duration
	private double  refeshDuration; // milliseconds
	
	// flag for appearance
	private boolean is3DBallShapeActivated;
	private boolean isMotionVectorActivated;
	
	// constructor
	public MovingBall(double radius, double refeshDuration,
					double minDepth, double maxDepth, 
					double screenWidth, double screenHeight, 
					boolean is3DBallShapeActivated, boolean isMotionVectorActivated) {
		super();	
		this.radius = radius;
		this.minDepth = minDepth;
		this.maxDepth = maxDepth;
		this.scrWidth = screenWidth;
		this.scrHeight= screenHeight;
		this.refeshDuration = refeshDuration;
		this.is3DBallShapeActivated = is3DBallShapeActivated;
		this.isMotionVectorActivated = isMotionVectorActivated;

		// initialize moving ball's real and onscreen positions
		// and accompanying motion vector parameters
		this.initBallAndMotionVectorPosition();
		
		// generate a ball	
		this.ball = new Ball(this.is3DBallShapeActivated);
		this.ball.setBall(this.radius, this.pixX, this.pixY);
		
		// generate corresponding motion vector
		this.mv = new MotionVector(this.isMotionVectorActivated);
		this.mv.setMotionVector(this.fromPixX, this.fromPixY, this.toPixX, this.toPixY);		
		
		// add ball and vector 
		this.getChildren().add(this.ball);
		this.getChildren().add(this.mv);		
	}
	
	// place a ball at a random onscreen position and random distance
	// Math.rand() returns a pseudorandom double greater than or equal to 0.0 and less than 1.0
	public void initBallAndMotionVectorPosition() {
		
		// compute the random onscreen ball location	
		this.pixX = Math.random()*(this.scrWidth-this.radius)+1;
		this.pixY = Math.random()*(this.scrHeight-this.radius)+1;
				
		// update real-world position
		//this.currZ = this.maxDepth;
		this.currZ = this.minDepth+Math.random()*(this.maxDepth-this.minDepth);
		this.currX = Utils.getRealWorldCoordX(this.pixX, this.currZ);
		this.currY = Utils.getRealWorldCoordY(this.pixY, this.currZ);
		
		// initialize previous position for the first time
		this.prevX = this.currX;
		this.prevY = this.currY;
		this.prevZ = this.currZ;		
		
		this.fromPixX = Utils.getOnscreenCoordX(this.currX, this.currZ);
		this.fromPixY = Utils.getOnscreenCoordY(this.currY, this.currZ);
		this.toPixX   = Utils.getOnscreenCoordX(this.currX+(currX-prevX), this.currZ+(currZ-prevZ));
		this.toPixY   = Utils.getOnscreenCoordY(this.currY+(currY-prevY), this.currZ+(currZ-prevZ));
	}
	
	// update ball position based on user speed (displacement)
	public void move(double userLinearSpeedX, double userLinearSpeedY, double userLinearSpeedZ) {

		// keep previous real-world position for vector calculation
		this.prevX = this.currX;
		this.prevY = this.currY;
		this.prevZ = this.currZ;

		// compute new real-world position
		this.currX = this.currX-userLinearSpeedX*this.refeshDuration;
		this.currY = this.currY-userLinearSpeedY*this.refeshDuration;
		this.currZ = this.currZ-userLinearSpeedZ*this.refeshDuration;
				
		// transfer original position to distorted position
		this.currX = Trans.trnasferX(this.currX, this.currZ);
		this.currY = Trans.trnasferY(this.currY, this.currZ);
		this.currZ = Trans.trnasferZ(this.currZ);

		// compute new onscreen ball position
		this.pixX = Utils.getOnscreenCoordX(this.currX, this.currZ);
		this.pixY = Utils.getOnscreenCoordY(this.currY, this.currZ);		
		this.ball.setBall(this.radius, this.pixX, this.pixY);
		
		// update motion vector
		this.fromPixX = Utils.getOnscreenCoordX(this.currX, this.currZ);
		this.fromPixY = Utils.getOnscreenCoordY(this.currY, this.currZ);
		this.toPixX   = Utils.getOnscreenCoordX(this.currX+(currX-prevX), this.currZ+(currZ-prevZ));
		this.toPixY   = Utils.getOnscreenCoordY(this.currY+(currY-prevY), this.currZ+(currZ-prevZ));	
		this.mv.setMotionVector(fromPixX, fromPixY, toPixX, toPixY);		
		
		// if ball has been passed or off-screen then reuse it by reinitialize it
		if (!this.isBallVisible()) {
			this.initBallAndMotionVectorPosition();
		} 
	}
	
	// checker for ball visibility
	public boolean isBallVisible() {
		return !(currZ<this.minDepth || currZ>this.maxDepth
				|| this.pixX<0 || this.pixY<0	
				|| this.pixX>this.scrWidth || this.pixY>this.scrHeight);
	}
	
	// compute motion vector speed in m/s 
	private double motionVectorSize() {
		return Math.sqrt((currX-prevX)*(currX-prevX)
						+(currY-prevY)*(currY-prevY)
						+(currZ-prevZ)*(currZ-prevZ))/this.refeshDuration*1000;
	}
	
	
	
	// print moving ball 
	public String printBall() {
		return "PrevPos: "+ this.prevX+"\t"+this.prevY+"\t"+ this.prevZ+"\n"
				+"CurrPos: "+ this.currX+"\t"+this.currY+"\t"+this.currZ
				+"FromPix: "+ this.fromPixX+"\t"+this.fromPixY+"\n"
				+"ToPix: "+ this.toPixX+"\t"+this.toPixY+motionVectorSize()+"\n";
	}

}
