package application;
import javafx.application.*;
import javafx.stage.*;
import javafx.scene.*;
import javafx.scene.input.KeyCode;
import javafx.scene.paint.Color;
import javafx.animation.*;
import javafx.util.*;
import java.util.*;

public class SpaceDistortionTest extends Application {

    final private int BALL_SIZE 	= 3;	// flying ball size
    final private int N_BALLS 		= 1000;	// number of balls
    
    final private boolean IS_MOTION_VECTOR_ACTIVATED = true;
    final private boolean IS_3D_BALL_SHAPE_ACTIVATED = false;
    
    // parameter for parameter mismatch
    final private double kS = 1.0; // the ratio of eye separation to camera separation
    final private double kW = 1.0; // the ratio of screen width to camera frustum width at convergence distance,
    final private double kD = 1.0; // the ratio of screen distance to camera convergence distance       
    final private double dC = 3.0; // camera convergence distance (meters)

    // viewer moving speed along X, Y, and Z-directions in meter/milliseconds
    final private double USER_LINEAR_SPEED_X = 0.0/1000; 
    final private double USER_LINEAR_SPEED_Y = 0.0/1000; 
    final private double USER_LINEAR_SPEED_Z = 5.0/1000;
    
    // viewer angular speed (yaw, pitch, roll) in radian/milliseconds
    //final private double USER_SPEED_YAW 	= 0.0/1000; 
    //final private double USER_SPEED_PITCH = 0.0/1000; 
    //final private double USER_SPEED_ROLL 	= 1.0/1000;
    
    
    
    // animation parameter
    final private double REFRESH_RATE 	= 60.0; // Hz
    final private double REFRESH_DURATION = 1.0/REFRESH_RATE*1000; // milliseconds
    
    // visibility range for animation
    final private double MIN_DEPTH = 0.5; // meter
    final private double MAX_DEPTH = 10.0;// meter
    
    // screen and display sizes & viewer distance to screen
    final private double SCREEN_WIDTH 	= 1920; 	// pixels
    final private double SCREEN_HEIGHT 	= 1080; 	// pixels    
    final private double DISPLAY_WIDTH 	= 0.51; 	// meter
    final private double DISPLAY_HEIGHT = 0.29; 	// meter
    final private double DISTANCE_TO_DISPLAY= 0.60; // meter
            
    // background color
    final private Color  BACKGROUND_COLOR = Color.BLACK;
    
    // list of balls and strokes
    public static ArrayList<MovingBall> movingBalls = new ArrayList<MovingBall>();
    public static ArrayList<ReferenceLine> refLines = new ArrayList<ReferenceLine>();

    // main animation function
    @Override 
    public void start(final Stage primaryStage) {
    	
    	// initialize Transfer function
    	Trans.init(kS, kW, kD, dC);    
    	
    	// initialize Utility function
    	Utils.init(DISTANCE_TO_DISPLAY, DISPLAY_WIDTH, DISPLAY_HEIGHT, SCREEN_WIDTH, SCREEN_HEIGHT);    	
    	
        // generate balls with random location and generate motion vector
		for (int i = 0; i < N_BALLS; i++) {
			movingBalls.add(
					new MovingBall(BALL_SIZE, REFRESH_DURATION, 
									MIN_DEPTH, MAX_DEPTH, 
									SCREEN_WIDTH, SCREEN_HEIGHT, 
									IS_3D_BALL_SHAPE_ACTIVATED, IS_MOTION_VECTOR_ACTIVATED)
			);
		}

		// generate a horizontal and vertical lines
		refLines.add(new ReferenceLine(0, SCREEN_HEIGHT/2, SCREEN_WIDTH,SCREEN_HEIGHT/2));
		refLines.add(new ReferenceLine(SCREEN_WIDTH/2, 0, SCREEN_WIDTH/2,SCREEN_HEIGHT));
		
    	// set drawing elements
        Group root = new Group();
        		
		// add rendering nodes to root
		root.getChildren().add(new Background(SCREEN_WIDTH, SCREEN_HEIGHT, BACKGROUND_COLOR));
		root.getChildren().addAll(refLines);
 		root.getChildren().addAll(movingBalls);
 		
 		// create a scene
        Scene scene = this.generateScene(root);   
                       
        // finalize the view
        primaryStage.setTitle("Optic flow");
        primaryStage.setScene(scene);
        primaryStage.setFullScreenExitHint("");
        primaryStage.setFullScreen(true);
        primaryStage.show();

		// set up time line and play 
        Timeline t = new Timeline(this.generateKeyFrame());
        t.setCycleCount(Timeline.INDEFINITE);
        t.play();
    }

    // Generate a key frame for every refresh timing
    private KeyFrame generateKeyFrame() {
    	return new KeyFrame(Duration.millis(REFRESH_DURATION), event -> {
			// recalculate ball position and motion vector
			for (int i=0;i<movingBalls.size();i++) {
				
				// update ball position				
				movingBalls.get(i).move(USER_LINEAR_SPEED_X, USER_LINEAR_SPEED_Y, USER_LINEAR_SPEED_Z);
			}
		});
    }

    // Generate a scene
    private Scene generateScene(Group root) {
    	Scene scene = new Scene(root, SCREEN_WIDTH, SCREEN_HEIGHT);
        
        //Add a Movement Keyboard Handler
        scene.setOnKeyPressed( event -> {
            KeyCode keycode = event.getCode();
            
            // if Q or ESC pressed quit
            if(keycode == KeyCode.Q || keycode == KeyCode.ESCAPE) { 
            	System.exit(0); 
            }
        });
    	return scene;
    }

	// Main program for launcher
    public static void main(String[] args) {
        launch(args);
    }
}